/***********************************************************************
 *
 * Copyright (C) 2007 Graeme Gott <gottcode@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 **********************************************************************/

#include "board.h"

#include "maze.h"
#include "theme.h"

#include <QKeyEvent>
#include <QMainWindow>
#include <QMessageBox>
#include <QPainter>
#include <QSettings>
#include <QStatusBar>
#include <QSvgRenderer>

// ============================================================================

Board::Board(QMainWindow* parent)
:	QLabel(parent),
	m_seed(0),
	m_total_targets(3),
	m_maze(0),
	m_player_angle(360)
{
	setMinimumSize(448, 448);
	setAlignment(Qt::AlignCenter);

	// Create status message
	m_status_message = new QLabel;
	parent->statusBar()->addPermanentWidget(m_status_message);

	// Setup theme support
	m_theme = new Theme;

	// Start or load game
	m_seed = QSettings().value("Current/Seed", 0).toUInt();
	if (m_seed) {
		loadGame();
	} else {
		newGame();
	}

	loadSettings();
}

// ============================================================================

Board::~Board()
{
	delete m_maze;
	delete m_theme;
}

// ============================================================================

void Board::newGame()
{
	// Prompt user
	if (m_seed) {
		if (QMessageBox::question(this, tr("CuteMaze"), tr("Abort current game?"), QMessageBox::Yes | QMessageBox::No, QMessageBox::Yes) == QMessageBox::No) {
			return;
		}
	}

	// Fetch new seed
	srand(time(0));
	do {
		m_seed = rand();
	} while (m_seed == 0);

	// Set values for new game
	QSettings settings;
	settings.remove("Current");
	settings.setValue("Current/Seed", m_seed);
	m_total_targets = settings.value("New/Targets", 3).toInt();
	m_total_targets = m_total_targets > 0 ? m_total_targets : 1;
	m_total_targets = m_total_targets < 11 ? m_total_targets : 10;
	settings.setValue("Current/Targets", m_total_targets);
	settings.setValue("Current/Size", settings.value("New/Size", 50).toInt());
	settings.setValue("Current/Algorithm", settings.value("New/Algorithm", 4).toInt());
	m_player_angle = 360;

	// Create new game
	generate();
	saveGame();

	// Show
	setPixmap(render());
}

// ============================================================================

void Board::loadGame()
{
	m_status_message->clear();

	QSettings settings;

	// Load maze
	m_total_targets = settings.value("Current/Targets", 3).toInt();
	m_total_targets = m_total_targets > 0 ? m_total_targets : 1;
	m_total_targets = m_total_targets < 11 ? m_total_targets : 10;
	generate();
	if (!m_maze->load()) {
		QMessageBox::warning(this, tr("CuteMaze"), tr("Unable to load previous game. A new game will be started."));
		m_seed = 0;
		return newGame();
	}

	// Remove any targets with matching movement
	for (int i = 0; i < m_targets.size(); ++i) {
		if (m_maze->cell(m_targets[i].x(), m_targets[i].y()).pathMarker()) {
			m_targets.takeAt(i);
			--i;
		}
	}

	// Set status message
	updateTargetMessage();

	// Place player at last location
	m_player = settings.value("Current/Player").toPoint();
	m_player_angle = settings.value("Current/Rotation", 360).toInt();
	if (m_player_angle % 90 != 0 || m_player_angle < 0 || m_player_angle > 360) {
		m_player_angle = 360;
	}

	// Show
	setPixmap(render());

	// Should not happen, but handle a finished game
	if (m_targets.isEmpty()) {
		finish();
	}
}

// ============================================================================

void Board::saveGame()
{
	if (m_seed) {
		m_maze->save();
		QSettings().setValue("Current/Player", m_player);
		QSettings().setValue("Current/Rotation", m_player_angle);
	}
}

// ============================================================================

void Board::loadSettings()
{
	QSettings settings;

	// Load theme
	m_theme->load(settings.value("Theme", "Mouse").toString());
	if (m_seed) {
		setPixmap(render());
	} else {
		setPixmap(renderDone());
	}

	// Load player controls
	m_controls_up = settings.value("Controls/Up", Qt::Key_Up).toUInt();
	m_controls_down = settings.value("Controls/Down", Qt::Key_Down).toUInt();
	m_controls_left = settings.value("Controls/Left", Qt::Key_Left).toUInt();
	m_controls_right = settings.value("Controls/Right", Qt::Key_Right).toUInt();
	m_controls_flag = settings.value("Controls/Flag", Qt::Key_Space).toUInt();
}

// ============================================================================

void Board::keyPressEvent(QKeyEvent* event)
{
	// Prevent player from changing a finished maze
	if (!m_seed) {
		return;
	}

	QPoint position = m_player;
	const Cell& cell = m_maze->cell(m_player.x(), m_player.y());

	unsigned int keypress = event->key();
	if (keypress == m_controls_left) {
		m_player_angle = 270;
		if (!cell.leftWall()) {
			Q_ASSERT(m_player.x() > 0);
			m_player.rx()--;
		}
	} else if (keypress == m_controls_right) {
		m_player_angle = 90;
		if (!cell.rightWall()) {
			Q_ASSERT(m_player.x() < m_maze->columns() - 1);
			m_player.rx()++;
		}
	} else if (keypress == m_controls_up) {
		m_player_angle = 360;
		if (!cell.topWall()) {
			Q_ASSERT(m_player.y() > 0);
			m_player.ry()--;
		}
	} else if (keypress == m_controls_down) {
		m_player_angle = 180;
		if (!cell.bottomWall()) {
			Q_ASSERT(m_player.y() < m_maze->rows() - 1);
			m_player.ry()++;
		}
	} else if (keypress == m_controls_flag) {
		m_maze->cellMutable(m_player.x(), m_player.y()).toggleFlag();
	} else {
		return;
	}

	// Add path marker
	if ( (position != m_player) && (m_maze->cell(position.x(), position.y()).pathMarker() == 0) ) {
		int cdelta = m_player.x() - position.x();
		int rdelta = m_player.y() - position.y();
		int angle = 0;
		if (cdelta) {
			angle = 180 - (cdelta * 90);
		} else {
			angle = 360 - ((rdelta + 1) * 90);
		}
		m_maze->cellMutable(position.x(), position.y()).setPathMarker(angle);
	}

	// Check for collisions with targets
	for (int i = 0; i < m_targets.size(); ++i) {
		if (m_player == m_targets.at(i)) {
			m_targets.takeAt(i);
			--i;
			updateTargetMessage();
		}
	}

	// Show updated maze
	setPixmap(render());

	// Handle finishing a maze
	if (m_targets.isEmpty()) {
		m_maze->cellMutable(m_player.x(), m_player.y()).setPathMarker(m_player_angle);
		finish();
	}
}

// ============================================================================

void Board::resizeEvent(QResizeEvent* event)
{
	QLabel::resizeEvent(event);
	if (m_seed) {
		setPixmap(render());
	} else {
		setPixmap(renderDone());
	}
}

// ============================================================================

void Board::generate()
{
	int columns = QSettings().value("Current/Size", 50).toInt();
	columns = columns > 9 ? columns : 10;
	columns = columns < 101 ? columns : 100;
	int rows = columns;

	// Create new maze
	m_targets.clear();
	srand(m_seed);
	delete m_maze;
	switch (QSettings().value("Current/Algorithm", 4).toInt()) {
	case 0:
		m_maze = new HuntAndKillMaze;
		break;
	case 1:
		m_maze = new KruskalMaze;
		break;
	case 2:
		m_maze = new PrimMaze;
		break;
	case 3:
		m_maze = new RecursiveBacktrackerMaze;
		break;
	case 5:
		m_maze = new Stack2Maze;
		break;
	case 6:
		m_maze = new Stack3Maze;
		break;
	case 7:
		m_maze = new Stack4Maze;
		break;
	case 8:
		m_maze = new Stack5Maze;
		break;
	case 4:
	default:
		m_maze = new StackMaze;
		break;
	}
	m_maze->generate(columns, rows);

	// Add player
	m_player.setX(rand() % (columns - 1));
	m_player.setY(rand() % (rows - 1));
	m_start = m_player;

	// Add targets
	QList<QPoint> locations;
	locations.append(m_start);
	for (int i = 0; i < m_total_targets; ++i) {
		QPoint target;
		do {
			target.setX(rand() % (columns - 1));
			target.setY(rand() % (rows - 1));
		} while (locations.contains(target));
		m_targets.append(target);
	}
	updateTargetMessage();
}

// ============================================================================

void Board::finish()
{
	// Remove game from disk
	m_seed = 0;
	QSettings().remove("Current");

	// Show congratulations
	m_status_message->setText(tr("Success!"));
	setPixmap(renderDone());
}

// ============================================================================

void Board::updateTargetMessage()
{
	m_status_message->setText(tr("%1 of %2 targets remaining").arg(m_targets.size()).arg(m_total_targets));
}

// ============================================================================

QPixmap Board::render() const
{
	int column = m_player.x() - 2;
	int row = m_player.y() - 2;
	int columns = m_maze->columns();
	int rows = m_maze->rows();

	Q_ASSERT(column > -3);
	Q_ASSERT(column < columns - 2);
	Q_ASSERT(row > -3);
	Q_ASSERT(row < rows - 2);

	// Create pixmap
	int size = qMin(width(), height());
	size -= (size % 14);
	QPixmap pixmap(size, size);
	pixmap.fill(Qt::white);

	// Create painter
	QPainter painter(&pixmap);
	float scale = static_cast<float>(size) / 448.0f;
	painter.scale(scale, scale);

	// Draw background
	for (int r = 0; r < 6; ++r) {
		for (int c = 0; c < 6; ++c) {
			m_theme->draw(painter, c, r, Theme::Background);
		}
	}

	// Initialize corners
	unsigned char corners[6][6];
	for (int r = 0; r < 6; ++r) {
		for (int c = 0; c < 6; ++c) {
			corners[c][r] = 0;
		}
	}

	// Setup columns
	int column_start = 0;
	int column_count = 5;
	if (column < 1) {
		column_start = abs(column);
	} else if (column + 4 >= columns) {
		column_count = columns - column;
	}

	// Setup rows
	int row_start = 0;
	int row_count = 5;
	if (row < 1) {
		row_start = abs(row);
	} else if (row + 4 >= rows) {
		row_count = rows - row;
	}

	// Draw cells
	int angle = 0;
	for (int r = row_start; r < row_count; ++r) {
		for (int c = column_start; c < column_count; ++c) {

			const Cell& cell = m_maze->cell(column + c, row + r);

			// Draw walls
			if (cell.topWall()) {
				m_theme->drawWall(painter, c, r);
			}
			if (cell.leftWall()) {
				m_theme->drawWall(painter, c, r, true);
			}
			if (column + c + 1 == columns) {
				m_theme->drawWall(painter, c + 1, r, true);
			}
			if (row + r + 1 == rows) {
				m_theme->drawWall(painter, c, r + 1);
			}

			// Draw marker
			angle = cell.pathMarker();
			if (angle) {
				m_theme->draw(painter, c, r, Theme::Marker, angle);
			}

			// Draw flag
			if (cell.flag()) {
				m_theme->draw(painter, c, r, Theme::Flag);
			}

			// Configure corners
			unsigned char& corner1 = corners[c][r];
			corner1 |= (cell.topWall() << 1);
			corner1 |= (cell.leftWall() << 2);
			unsigned char& corner2 = corners[c + 1][r];
			corner2 |= (cell.topWall() << 3);
			corner2 |= (cell.rightWall() << 2);
			unsigned char& corner3 = corners[c + 1][r + 1];
			corner3 |= (cell.rightWall() << 0);
			corner3 |= (cell.bottomWall() << 3);
			unsigned char& corner4 = corners[c][r + 1];
			corner4 |= (cell.leftWall() << 0);
			corner4 |= (cell.bottomWall() << 1);
		}
	}

	// Draw corners
	for (int r = 0; r < 6; ++r) {
		for (int c = 0; c < 6; ++c) {
			unsigned char walls = corners[c][r];
			if (walls) {
				m_theme->drawCorner(painter, c, r, walls);
			}
		}
	}

	// Draw start
	QRect view(column, row, 5, 5);
	if (view.contains(m_start)) {
		m_theme->draw(painter, m_start.x() - column, m_start.y() - row, Theme::Start);
	}

	// Draw targets
	foreach (QPoint target, m_targets) {
		if (view.contains(target)) {
			m_theme->draw(painter, target.x() - column, target.y() - row, Theme::Target);
		}
	}

	// Draw player
	m_theme->draw(painter, 2, 2, Theme::Player, m_player_angle);

	return pixmap;
}

// ============================================================================

QPixmap Board::renderDone() const
{
	int columns = m_maze->columns();
	int rows = m_maze->rows();

	// Determine cell width
	int mcr = qMin(columns, rows);
	int cell_width = qMin(width(), height());
	cell_width -= (mcr + 1);
	cell_width /= mcr;
	cell_width += 1;

	// Create pixmap
	QPixmap pixmap(columns * cell_width + 1, rows * cell_width + 1);
	pixmap.fill(Qt::white);

	// Draw image
	QPainter painter(&pixmap);
	int x1, x2, y1, y2;
	for (int r = 0; r < rows; ++r) {
		for (int c = 0; c < columns; ++c) {
			const Cell& cell = m_maze->cell(c, r);
			x1 = c  * cell_width;
			x2 = x1 + cell_width;
			y1 = r  * cell_width;
			y2 = y1 + cell_width;
			if (cell.pathMarker()) {
				painter.fillRect(x1 + 1, y1 + 1, cell_width, cell_width, Qt::lightGray);
			}
			if (cell.topWall()) {
				painter.drawLine(x1, y1, x2, y1);
			}
			if (cell.leftWall()) {
				painter.drawLine(x1, y1, x1, y2);
			}
		}
	}
	painter.drawLine(0, rows * cell_width, columns * cell_width, rows * cell_width);
	painter.drawLine(columns * cell_width, 0, columns * cell_width, rows * cell_width);

	// Draw congratulations
	QString message(QT_TR_NOOP("Success!"));
	painter.setFont(QFont("Sans", 24));
	QRect rect = painter.fontMetrics().boundingRect(message);
	x1 = (pixmap.width() >> 1) - (rect.width() + rect.height() >> 1);
	y1 = (pixmap.height() >> 1) - (rect.height() * 2 >> 1);
	painter.setPen(Qt::NoPen);
	painter.setBrush(QColor(0, 0, 0, 200));
	painter.setRenderHint(QPainter::Antialiasing, true);
	painter.drawRoundRect(x1, y1, rect.width() + rect.height(), rect.height() * 2, 10);
	painter.setPen(Qt::white);
	painter.setRenderHint(QPainter::TextAntialiasing, true);
	painter.drawText(rect.height() / 2 + x1, rect.height() * 1.25 + y1, message);

	return pixmap;
}

// ============================================================================
