/***********************************************************************
 *
 * Copyright (C) 2007 Graeme Gott <gottcode@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 **********************************************************************/

#include "maze.h"

#include <QSettings>

// Helpers
// ============================================================================

namespace
{

QPoint direction[4] = {
	QPoint(1, 0),
	QPoint(0, 1),
	QPoint(-1, 0),
	QPoint(0, -1)
};

QPoint randomNeighbor(QList<QPoint>& unvisited, const QPoint& cell)
{
	// Find unvisited neighbors
	QPoint n1 = cell + direction[0];
	QPoint n2 = cell + direction[1];
	QPoint n3 = cell + direction[2];
	QPoint n4 = cell + direction[3];
	int neighbors[4] = { -1, -1, -1, 1 };
	int found = 0;
	int size = unvisited.size();
	for (int i = 0; i < size; ++i) {
		const QPoint& cell = unvisited.at(i);
		if (n1 == cell || n2 == cell || n3 == cell || n4 == cell) {
			neighbors[found] = i;
			found++;
		}
	}

	// Return random neighbor
	if (found) {
		return unvisited.takeAt(neighbors[rand() % found]);
	} else {
		return QPoint(-1,-1);
	}
}

}

// ============================================================================

// Hunt and Kill algorithm
// ============================================================================

void HuntAndKillMaze::generate()
{
	// Generate unvisited list
	m_unvisited.clear();
	for (int c = 0; c < columns(); ++c) {
		for (int r = 0; r < rows(); ++r) {
			m_unvisited.append(QPoint(c, r));
		}
	}
	QPoint current = m_unvisited.takeAt(rand() % columns());

	QPoint neighbor;
	while (!m_unvisited.isEmpty()) {
		neighbor = randomNeighbor(m_unvisited, current);
		if (neighbor.x() != -1) {
			mergeCells(current, neighbor);
			current = neighbor;
		} else {
			current = hunt();
		}
	}
}

// ============================================================================

QPoint HuntAndKillMaze::hunt()
{
	QPoint cell, next;
	int size = m_unvisited.size();
	for (int i = 0; i < size; ++i) {
		cell = m_unvisited.at(i);
		for (int d = 0; d < 4; ++d) {
			next = cell + direction[d];
			if (next.x() < 0 ||
				next.x() >= columns() ||
				next.y() < 0 ||
				next.y() >= rows()) {
				continue;
			}
			if (!m_unvisited.contains(next)) {
				mergeCells(cell, next);
				m_unvisited.takeAt(i);
				return cell;
			}
		}
	}
	return cell;
}

// ============================================================================

// Kruskal's algorithm
// ============================================================================

void KruskalMaze::generate()
{
	// Generate sets
	m_sets.clear();
	for (int c = 0; c < columns(); ++c) {
		for (int r = 0; r < rows(); ++r) {
			m_sets.append(QList<QPoint>() << QPoint(c, r));
		}
	}

	while (m_sets.size() > 1) {
		int set1 = 0;
		int set2 = -1;

		// Find random cell
		QPoint cell = m_sets.at(set1).at(rand() % m_sets.at(set1).size());
		int c = cell.x();
		int r = cell.y();

		// Find random neighbor of cell
		if (rand() % 2) {
			c += 1;
		} else {
			r += 1;
		}
		QPoint cell2(c, r);
		if (cell2.x() >= columns() || cell2.y() >= rows()) {
			continue;
		}

		// Find set containing second cell
		int size = m_sets.size();
		for (int i = 0; i < size; ++i) {
			if (m_sets.at(i).contains(cell2)) {
				set2 = i;
				break;
			}
		}

		// Merge sets if they are different
		if (set1 != set2) {
			mergeCells(cell, cell2);
			m_sets[set2] += m_sets.at(set1);
			m_sets.takeAt(set1);
		}
	}
}

// ============================================================================

// Prim's algorithm
// ============================================================================

void PrimMaze::generate()
{
	// Generate cell lists
	m_in.clear();
	m_frontier.clear();
	m_out.clear();
	for (int c = 0; c < columns(); ++c) {
		for (int r = 0; r < rows(); ++r) {
			m_out.append(QPoint(c, r));
		}
	}

	// Move first cell
	QPoint cell = m_out.takeAt( rand() % columns() );
	m_in.append(cell);
	moveNeighbors(cell);

	// Move remaining cells
	while (!m_frontier.isEmpty()) {
		cell = m_frontier.takeAt( rand() % m_frontier.size() );
		mergeRandomNeighbor(cell);
		m_in.append(cell);
		moveNeighbors(cell);
	}
}

// ============================================================================

bool PrimMaze::areNeighbors(const QPoint& cell1, const QPoint& cell2) const
{
	int c1 = cell1.x();
	int r1 = cell1.y();
	int c2 = cell2.x();
	int r2 = cell2.y();

	if (r1 == r2) {
		if (c1 == c2 + 1 || c1 == c2 - 1) {
			return true;
		}
	} else if (c1 == c2) {
		if (r1 == r2 + 1 || r1 == r2 - 1) {
			return true;
		}
	}
	return false;
}

// ============================================================================

void PrimMaze::moveNeighbors(const QPoint& cell)
{
	int size = m_out.size();
	for (int i = 0; i < size; ++i) {
		if (areNeighbors(cell, m_out.at(i))) {
			m_frontier.append(m_out.takeAt(i));
			--size;
		}
	}
}

// ============================================================================

void PrimMaze::mergeRandomNeighbor(const QPoint& cell)
{
	QList<int> neighbors;
	int size = m_in.size();
	for (int i = 0; i < size; ++i) {
		if (areNeighbors(cell, m_in.at(i))) {
			neighbors.append(i);
		}
	}
	QPoint neighbor = m_in[neighbors[rand() % neighbors.size()]];
	mergeCells(cell, neighbor);
}

// ============================================================================

// Recursive Backtracker algorithm
// ============================================================================

void RecursiveBacktrackerMaze::generate()
{
	// Generate unvisited list
	m_unvisited.clear();
	for (int c = 0; c < columns(); ++c) {
		for (int r = 0; r < rows(); ++r) {
			m_unvisited.append(QPoint(c, r));
		}
	}
	makePath( m_unvisited.takeAt(rand() % columns()) );
}

// ============================================================================

void RecursiveBacktrackerMaze::makePath(const QPoint& current)
{
	QPoint neighbor;
	while ( (neighbor = randomNeighbor(m_unvisited, current)).x() != -1 ) {
		mergeCells(current, neighbor);
		makePath(neighbor);
	}
}

// ============================================================================

// Stack algorithm
// ============================================================================

void StackMaze::generate()
{
	// Generate cell lists
	m_unvisited.clear();
	for (int c = 0; c < columns(); ++c) {
		for (int r = 0; r < rows(); ++r) {
			m_unvisited.append(QPoint(c, r));
		}
	}
	QList<QPoint> active;
	active.append(m_unvisited.takeAt(rand() % columns()));

	// Loop through active list
	QPoint cell, neighbor;
	int pos;
	while (!active.isEmpty()) {
		pos = nextActive(active.size());
		cell = active.at(pos);
		neighbor = randomNeighbor(m_unvisited, cell);
		if (neighbor.x() != -1) {
			mergeCells(cell, neighbor);
			active.append(neighbor);
		} else {
			active.takeAt(pos);
		}
	}
}

// ============================================================================

int StackMaze::nextActive(int size)
{
	return size - 1;
}

// ============================================================================

int Stack2Maze::nextActive(int size)
{
	if (rand() % 2 != 0) {
		return size - 1;
	} else {
		return rand() % (size);
	}
}

// ============================================================================

int Stack3Maze::nextActive(int size)
{
	return rand() % size;
}

// ============================================================================

int Stack4Maze::nextActive(int size)
{
	int recent = 3 < size ? 3 : size;
	return size - (rand() % recent) - 1;
}

// ============================================================================

int Stack5Maze::nextActive(int size)
{
	switch (rand() % 3) {
	case 0:
		return 0;
	case 1:
		return rand() % size;
	case 2:
	default:
		return size - 1;
	}
}

// ============================================================================

// Maze class
// ============================================================================

void Maze::generate(int columns, int rows)
{
	m_columns = columns;
	m_rows = rows;
	m_cells = QVector< QVector<Cell> >(m_columns, QVector<Cell>(m_rows));
	generate();
}

// ============================================================================

bool Maze::load()
{
	// Read data from disk
	QByteArray data = QSettings().value("Current/Progress").toByteArray();
	if (data.isEmpty()) {
		return false;
	}

	// Decompress data
	data = qUncompress(data);
	if (data.isEmpty()) {
		return false;
	}

	// Deserialize data
	QDataStream stream(&data, QIODevice::ReadOnly);
	stream.setVersion(QDataStream::Qt_4_3);
	for (int c = 0; c < m_columns; ++c) {
		for (int r = 0; r < m_rows; ++r) {
			stream >> m_cells[c][r];
		}
	}

	return true;
}

// ============================================================================

void Maze::save() const
{
	// Serialize data
	QByteArray data;
	QDataStream stream(&data, QIODevice::WriteOnly);
	stream.setVersion(QDataStream::Qt_4_3);
	for (int c = 0; c < m_columns; ++c) {
		for (int r = 0; r < m_rows; ++r) {
			stream << m_cells[c][r];
		}
	}

	// Compress data
	data = qCompress(data, 9);

	// Write data to disk
	QSettings().setValue("Current/Progress", data);
}

// ============================================================================

void Maze::mergeCells(const QPoint& cell1, const QPoint& cell2)
{
	if (cell1.y() == cell2.y()) {
		if (cell2.x() > cell1.x()) {
			m_cells[cell1.x()][cell1.y()].removeRightWall();
			m_cells[cell2.x()][cell2.y()].removeLeftWall();
		} else if (cell2.x() < cell1.x()) {
			m_cells[cell1.x()][cell1.y()].removeLeftWall();
			m_cells[cell2.x()][cell2.y()].removeRightWall();
		}
	} else if (cell1.x() == cell2.x()) {
		if (cell2.y() > cell1.y()) {
			m_cells[cell1.x()][cell1.y()].removeBottomWall();
			m_cells[cell2.x()][cell2.y()].removeTopWall();
		} else if (cell2.y() < cell1.y()) {
			m_cells[cell1.x()][cell1.y()].removeTopWall();
			m_cells[cell2.x()][cell2.y()].removeBottomWall();
		}
	}
}

// ============================================================================
