/***********************************************************************
 *
 * Copyright (C) 2007 Graeme Gott <gottcode@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 **********************************************************************/

#include "settings.h"

#include "theme.h"

#include <QComboBox>
#include <QDialogButtonBox>
#include <QGridLayout>
#include <QHBoxLayout>
#include <QKeyEvent>
#include <QLabel>
#include <QListWidget>
#include <QPainter>
#include <QPushButton>
#include <QSettings>
#include <QSpinBox>
#include <QTabWidget>
#include <QVBoxLayout>

// ControlButton class
//namespace
//{
// ============================================================================

class ControlButton : public QPushButton
{
public:
	ControlButton(const QString& type, Qt::Key default_key_value, Settings* parent);

	unsigned int key;
	unsigned int default_key;

protected:
	virtual void keyPressEvent(QKeyEvent* event);
	virtual void mousePressEvent(QMouseEvent* event);

private:
	Settings* m_settings;
};

// ============================================================================

ControlButton::ControlButton(const QString& type, Qt::Key default_key_value, Settings* parent)
:	QPushButton(parent),
	key(default_key_value),
	default_key(default_key_value),
	m_settings(parent)
{
	setObjectName("control_" + type);
	setText(QKeySequence(key).toString());
	setCheckable(true);
	setAutoDefault(false);
	installEventFilter(this);
}

// ============================================================================

void ControlButton::keyPressEvent(QKeyEvent* event)
{
	if ( (m_settings->m_active_control == this) && (event->key() != Qt::Key_Escape) &&  (event->count() == 1) ) {
		QString control = QKeySequence(event->key()).toString();
		foreach (ControlButton* button, m_settings->m_controls) {
			if (button->text() == control) {
				button->setText("");
			}
		}
		key = event->key();
		setText(control);
		setChecked(false);
		m_settings->m_active_control = 0;
	} else {
		QPushButton::keyPressEvent(event);
	}
}

// ============================================================================

void ControlButton::mousePressEvent(QMouseEvent* event)
{
	if (this != m_settings->m_active_control) {
		if (m_settings->m_active_control) {
			m_settings->m_active_control->setChecked(false);
		}
		m_settings->m_active_control = this;
	} else {
		m_settings->m_active_control = 0;
	}
	QPushButton::mousePressEvent(event);
}

// ============================================================================
//}

// Settings class
// ============================================================================

Settings::Settings(QWidget* parent)
:	QDialog(parent),
	m_active_control(0)
{
	setWindowTitle(tr("Settings"));
	m_theme = new Theme;

	QTabWidget* tabs = new QTabWidget(this);
	QDialogButtonBox* buttons = new QDialogButtonBox(QDialogButtonBox::Ok | QDialogButtonBox::Cancel, Qt::Horizontal, this);
	connect(buttons, SIGNAL(accepted()), this, SLOT(accept()));
	connect(buttons, SIGNAL(rejected()), this, SLOT(reject()));

	QVBoxLayout* layout = new QVBoxLayout(this);
	layout->setMargin(10);
	layout->setSpacing(18);
	layout->addWidget(tabs);
	layout->addWidget(buttons);

	resize(minimumSize());


	// Create Mazes tab
	QWidget* mazes_tab = new QWidget;
	tabs->addTab(mazes_tab, tr("Mazes"));

	m_mazes_algorithm = new QComboBox(mazes_tab);
	m_mazes_algorithm->setInsertPolicy(QComboBox::InsertAlphabetically);
	m_mazes_algorithm->addItem(tr("Hunt and Kill"), 0);
	m_mazes_algorithm->addItem(tr("Kruskal"), 1);
	m_mazes_algorithm->addItem(tr("Prim"), 2);
	m_mazes_algorithm->addItem(tr("Recursive Backtracker"), 3);
	m_mazes_algorithm->addItem(tr("Stack"), 4);
	m_mazes_algorithm->addItem(tr("Stack 2"), 5);
	m_mazes_algorithm->addItem(tr("Stack 3"), 6);
	m_mazes_algorithm->addItem(tr("Stack 4"), 7);
	m_mazes_algorithm->addItem(tr("Stack 5"), 8);
	connect(m_mazes_algorithm, SIGNAL(currentIndexChanged(int)), this, SLOT(algorithmSelected(int)));

	m_algorithm_preview = new QLabel(mazes_tab);

	m_mazes_targets = new QSpinBox(mazes_tab);
	m_mazes_targets->setRange(1, 10);

	m_mazes_size = new QSpinBox(mazes_tab);
	m_mazes_size->setRange(10, 100);

	QGridLayout* mazes_layout = new QGridLayout(mazes_tab);
	mazes_layout->setSpacing(6);
	mazes_layout->setRowStretch(0, 1);
	mazes_layout->setRowStretch(5, 1);
	mazes_layout->setColumnStretch(0, 1);
	mazes_layout->setColumnStretch(3, 1);
	mazes_layout->addWidget(m_algorithm_preview, 1, 2);
	mazes_layout->addWidget(new QLabel(tr("Algorithm"), mazes_tab), 2, 1, Qt::AlignRight | Qt::AlignVCenter);
	mazes_layout->addWidget(m_mazes_algorithm, 2, 2);
	mazes_layout->addWidget(new QLabel(tr("Targets"), mazes_tab), 3, 1, Qt::AlignRight | Qt::AlignVCenter);
	mazes_layout->addWidget(m_mazes_targets, 3, 2);
	mazes_layout->addWidget(new QLabel(tr("Size"), mazes_tab), 4, 1, Qt::AlignRight | Qt::AlignVCenter);
	mazes_layout->addWidget(m_mazes_size, 4, 2);


	// Create Controls tab
	QWidget* controls_tab = new QWidget;
	tabs->addTab(controls_tab, tr("Controls"));

	m_controls.append(new ControlButton("Up", Qt::Key_Up, this));
	m_controls.append(new ControlButton("Down", Qt::Key_Down, this));
	m_controls.append(new ControlButton("Left", Qt::Key_Left, this));
	m_controls.append(new ControlButton("Right", Qt::Key_Right, this));
	m_controls.append(new ControlButton("Flag", Qt::Key_Space, this));

	QGridLayout * controls_layout = new QGridLayout(controls_tab);
	controls_layout->setSpacing(6);
	controls_layout->setRowStretch(0, 1);
	controls_layout->setRowStretch(6, 1);
	controls_layout->setColumnStretch(0, 1);
	controls_layout->setColumnStretch(3, 1);
	controls_layout->addWidget(new QLabel(tr("Move Up"), controls_tab), 1, 1, Qt::AlignRight | Qt::AlignVCenter);
	controls_layout->addWidget(m_controls[0], 1, 2);
	controls_layout->addWidget(new QLabel(tr("Move Down"), controls_tab), 2, 1, Qt::AlignRight | Qt::AlignVCenter);
	controls_layout->addWidget(m_controls[1], 2, 2);
	controls_layout->addWidget(new QLabel(tr("Move Left"), controls_tab), 3, 1, Qt::AlignRight | Qt::AlignVCenter);
	controls_layout->addWidget(m_controls[2], 3, 2);
	controls_layout->addWidget(new QLabel(tr("Move Right"), controls_tab), 4, 1, Qt::AlignRight | Qt::AlignVCenter);
	controls_layout->addWidget(m_controls[3], 4, 2);
	controls_layout->addWidget(new QLabel(tr("Toggle Flag"), controls_tab), 5, 1, Qt::AlignRight | Qt::AlignVCenter);
	controls_layout->addWidget(m_controls[4], 5, 2);


	// Create Mazes tab
	QWidget* themes_tab = new QWidget;
	tabs->addTab(themes_tab, tr("Themes"));

	m_themes_selector = new QListWidget(themes_tab);
	connect(m_themes_selector, SIGNAL(currentTextChanged(const QString&)), this, SLOT(themeSelected(const QString&)));

	m_themes_preview = new QLabel(themes_tab);

	QHBoxLayout* themes_layout = new QHBoxLayout(themes_tab);
	themes_layout->setSpacing(6);
	themes_layout->addWidget(m_themes_selector);
	themes_layout->addWidget(m_themes_preview);


	// Load current settings
	loadSettings();
}

// ============================================================================

Settings::~Settings()
{
	delete m_theme;
}

// ============================================================================

void Settings::accept()
{
	QSettings settings;

	// Write new maze settings to disk
	settings.setValue("New/Algorithm", m_mazes_algorithm->itemData(m_mazes_algorithm->currentIndex()));
	settings.setValue("New/Targets", m_mazes_targets->value());
	settings.setValue("New/Size", m_mazes_size->value());

	// Write control button settings to disk
	foreach (ControlButton* button, m_controls) {
		settings.setValue("Controls/" + button->objectName().mid(8), button->key);
	}

	// Write theme to disk
	settings.setValue("Theme", m_themes_selector->currentItem()->text());

	QDialog::accept();
}

// ============================================================================

void Settings::reject()
{
	loadSettings();
	QDialog::reject();
}

// ============================================================================

void Settings::algorithmSelected(int index)
{
	if (index != -1) {
		m_algorithm_preview->setPixmap( QString(":/preview%1.png").arg( m_mazes_algorithm->itemData(index).toInt()) );
	}
}

// ============================================================================

void Settings::themeSelected(const QString& theme)
{
	if (!theme.isEmpty()) {
		m_theme->load(theme);
		generatePreview();
	}
}

// ============================================================================

void Settings::loadSettings()
{
	QSettings settings;

	// Read new maze settings from disk
	int algorithm = settings.value("New/Algorithm", 4).toInt();
	m_mazes_algorithm->setCurrentIndex(m_mazes_algorithm->findData(algorithm));
	m_mazes_targets->setValue(settings.value("New/Targets", 3).toInt());
	m_mazes_size->setValue(settings.value("New/Size", 50).toInt());

	// Read control button settings from disk
	foreach (ControlButton* button, m_controls) {
		button->key = settings.value("Controls/" + button->objectName().mid(8), button->default_key).toUInt();
		button->setText(QKeySequence(button->key).toString());
	}

	// Read theme from disk
	QStringList themes = m_theme->available();
	int theme = themes.indexOf(settings.value("Theme", "Mouse").toString());
	if (theme == -1) {
		theme = themes.indexOf("Mouse");
	}
	m_themes_selector->clear();
	m_themes_selector->addItems(themes);
	m_themes_selector->setCurrentRow(theme);
}

// ============================================================================

void Settings::generatePreview()
{
	QPixmap pixmap(192, 192);
	pixmap.fill(Qt::white);
	{
		QPainter painter(&pixmap);
		painter.translate(-80, -80);
		for (int c = 0; c < 4; ++c) {
			for (int r = 0; r < 4; ++r) {
				m_theme->draw(painter, c, r, Theme::Background);
			}
		}

		m_theme->drawWall(painter, 1, 1);
		m_theme->drawWall(painter, 2, 1);
		m_theme->drawWall(painter, 2, 2);
		m_theme->drawWall(painter, 1, 3);
		m_theme->drawWall(painter, 2, 1, true);
		m_theme->drawWall(painter, 1, 2, true);
		m_theme->drawWall(painter, 3, 2, true);

		m_theme->drawCorner(painter, 1, 1, 10);
		m_theme->drawCorner(painter, 2, 1, 14);
		m_theme->drawCorner(painter, 3, 1, 10);
		m_theme->drawCorner(painter, 1, 2, 4);
		m_theme->drawCorner(painter, 2, 2, 3);
		m_theme->drawCorner(painter, 3, 2, 12);
		m_theme->drawCorner(painter, 1, 3, 3);
		m_theme->drawCorner(painter, 2, 3, 8);
		m_theme->drawCorner(painter, 3, 3, 5);

		m_theme->draw(painter, 1, 1, Theme::Start);
		m_theme->draw(painter, 1, 1, Theme::Marker, 180);
		m_theme->draw(painter, 1, 2, Theme::Marker, 90);
		m_theme->draw(painter, 1, 2, Theme::Flag);
		m_theme->draw(painter, 2, 1, Theme::Target);
		m_theme->draw(painter, 2, 2, Theme::Player);
	}

	m_themes_preview->setPixmap(pixmap);
}

// ============================================================================
