/***********************************************************************
 *
 * Copyright (C) 2007 Graeme Gott <gottcode@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 **********************************************************************/

#include "theme.h"

#include <QApplication>
#include <QDir>
#include <QFileInfo>
#include <QIcon>
#include <QMainWindow>
#include <QPainter>
#include <QSet>
#include <QSvgRenderer>

// ============================================================================

namespace
{

struct CornerType
{
	unsigned char renderer;
	unsigned char transform;
};
CornerType corner_types[15] = {
	{4, 1},
	{4, 2},
	{3, 0},
	{4, 3},
	{2, 1},
	{3, 1},
	{1, 3},
	{4, 0},
	{3, 3},
	{2, 0},
	{1, 2},
	{3, 2},
	{1, 1},
	{1, 0},
	{0, 0}
};
QTransform corner_transforms[4] = {
	QTransform(),
	QTransform().translate(32, 0).rotate(90),
	QTransform().translate(32, 32).rotate(180),
	QTransform().translate(0, 32).rotate(270)
};

}

// ============================================================================

Theme::Theme()
{
	m_svg[Background] = new QSvgRenderer;
	m_svg[Flag] = new QSvgRenderer;
	m_svg[Marker] = new QSvgRenderer;
	m_svg[Player] = new QSvgRenderer;
	m_svg[Start] = new QSvgRenderer;
	m_svg[Target] = new QSvgRenderer;
	for (int i = 0; i < 5; ++i) {
		m_svg_corner[i] = new QSvgRenderer;
	}
	m_svg_wall = new QSvgRenderer;

	// Load theme locations
	QString xdg;
	xdg = getenv("$XDG_DATA_HOME");
	if (xdg.isEmpty()) {
		xdg = QDir::homePath() + "/.local/share/";
	}
	QStringList locations = xdg.split(":");
	xdg = getenv("$XDG_DATA_DIRS");
	if (xdg.isEmpty()) {
		xdg = "/usr/local/share/:/usr/share/";
	}
	locations += xdg.split(":");
	locations += ":";
	for (int i = 0; i < locations.size(); ++i) {
		locations[i] += "/games/cutemaze";
	}
	QDir::setSearchPaths("themes", locations);

	load("Mouse");
}

// ============================================================================

Theme::~Theme()
{
	delete m_svg[Background];
	delete m_svg[Flag];
	delete m_svg[Marker];
	delete m_svg[Player];
	delete m_svg[Start];
	delete m_svg[Target];
	for (int i = 0; i < 5; ++i) {
		delete m_svg_corner[i];
	}
	delete m_svg_wall;
}

// ============================================================================

QStringList Theme::available() const
{
	QSet<QString> dirs;
	dirs.insert("Mouse");
	dirs.insert("Penguin");

	QDir dir;
	foreach (QString path, QDir::searchPaths("themes")) {
		if (dir.cd(path)) {
			dirs += dir.entryList(QDir::Dirs | QDir::NoDotAndDotDot).toSet();
		}
	}

	QStringList list = QStringList::fromSet(dirs);
	list.sort();
	return list;
}

// ============================================================================

void Theme::load(const QString& name)
{
	QIcon icon("themes:" + name + "/player.svg");
	qApp->setWindowIcon(icon);
	foreach (QWidget* window, qApp->topLevelWidgets()) {
		if (qobject_cast<QMainWindow*>(window)) {
			window->setWindowIcon(icon);
		}
	}
	m_svg[Background]->load(QString("themes:" + name + "/background.svg"));
	m_svg[Flag]->load(QString("themes:" + name + "/flag.svg"));
	m_svg[Marker]->load(QString("themes:" + name + "/marker.svg"));
	m_svg[Player]->load(QString("themes:" + name + "/player.svg"));
	m_svg[Start]->load(QString("themes:" + name + "/start.svg"));
	m_svg[Target]->load(QString("themes:" + name + "/target.svg"));
	for (int i = 0; i < 5; ++i) {
		m_svg_corner[i]->load(QString("themes:" + name + "/corner%1.svg").arg(i));
	}
	m_svg_wall->load(QString("themes:" + name + "/wall.svg"));
}

// ============================================================================

void Theme::draw(QPainter& painter, int column, int row, enum Element element) const
{
	Q_ASSERT(element != TotalElements);
	QRect bounds(0, 0, 64, 64);
	if (element == Background) {
		bounds = QRect(-32, -32, 96, 96);
	}
	painter.save();
	painter.translate(column * 96, row * 96);
	m_svg[element]->render(&painter, bounds);
	painter.restore();
}

// ============================================================================

void Theme::draw(QPainter& painter, int column, int row, enum Element element, int angle) const
{
	Q_ASSERT(angle == 90 || angle == 180 || angle == 270 || angle == 360);
	painter.save();
	painter.translate(column * 96 + 32, row * 96 + 32);
	painter.rotate(angle);
	painter.translate(-32, -32);
	m_svg[element]->render(&painter, QRect(0, 0, 64, 64));
	painter.restore();
}

// ============================================================================

void Theme::drawCorner(QPainter& painter, int column, int row, unsigned char walls) const
{
	Q_ASSERT(walls > 0);
	Q_ASSERT(walls < 16);
	const CornerType& corner = corner_types[walls - 1];
	painter.save();
	painter.translate(column * 96 - 32, row * 96 - 32);
	painter.setWorldTransform(corner_transforms[corner.transform], true);
	m_svg_corner[corner.renderer]->render(&painter, QRect(0, 0, 32, 32));
	painter.restore();
}

// ============================================================================

void Theme::drawWall(QPainter& painter, int column, int row, bool vertical) const
{
	painter.save();
	painter.translate(column * 96, row * 96 + -32);
	if (vertical) {
		painter.rotate(90);
		painter.translate(32, 0);
	}
	m_svg_wall->render(&painter, QRect(0, 0, 64, 32));
	painter.restore();
}

// ============================================================================
